# Selective SLH in practice
This is the artifact for the paper `Selective SLH in Practice`

## Contents

This artifact contains the following directories

  - `chacha20`: code related to the experiments described in §V: application to crypto software
  - `jasmin`: a modified version of the Jasmin compiler (a part of which is previous work from the Jasmin authors,
  an other part is a contribution of this paper, see below)
  - `libjade-main`: the original version of the LibJade library
    (development version at commit ece99a3bbd8ebd831f285da0c909daba1ce2972d)
  - `libjade-sslh`: the modified and secure version of the LibJade library (a contribution of this paper)

## Contributions
The main contributions in this artifact are
- Jasmin SCT type checker with three new primitives/intrinsics and ability to spill to `mmx` registers.
- High-Assurance cryptographic implementations resistant to Spectre V1 using Jasmin SCT checked compiler.

## Jasmin
Jasmin is a framework containing both language and compiler which enables one to implement High-Assurance cryptographic algorithms.

### Building compiler

The recommended way to build the modified Jasmin compiler is using nix; for installation
instructions of nix, see https://nixos.org/download.html
To build Jasmin using nix, run in the `jasmin/` directory, the command

    nix-shell --run 'make -j2'

Detailed build instruction can be found in the `jasmin/` directory.

Once the compiler is built, it is advised to put it in your path or to run, in this directory:

    export JASMIN=$PWD/jasmin/compiler/jasminc

### Remarks

`jasmin/compiler/src/sct_checker_forward.ml` file contains main implemention of the SCT type checker in Ocaml.

`jasmin/compiler/src/constraints.ml` file contains constraint generation.


## Benchmarking
`libjade-main` contains crypto algorithms implemented in Constant-Time.
`libjade-sslh` contains crypto algorithms implemented using selective SLH and is in Speculative(v1) Constant-Time.

```
make
```
It will compile above two implementations and generate a `main-sslh` folder containing performance overhead for different algorithms.

collect the generated results into latex macros
```
python3 gen_latex_results1_data.py --indir main-sslh --outfile results1_data.tex
python3 gen_latex_results2_data.py --indir main-sslh --outfile results2_data.tex
```


## ChaCha20

Prerequisite: build the Jasmin compiler and export the `JASMIN`  environment variable (see above)

Build the various ChaCha20 versions: `make -C chacha20/src`
Check their security: `make -C chacha20/src check-sct`
Run unit tests: `make -C chacha20/test`
Run benchmarks: `make -C chacha20/bench`

## Secure LibJade

Prerequisite: build the Jasmin compiler and export the `JASMIN`  environment variable (see above)

Build the whole library: `make -C libjade-sslh/src`
Check its security: `make -C libjade-sslh/src check-sct`

