#!/usr/bin/env python3

import os
import csv
import sys
import argparse

impls1 = {
            "anon_kem_kyber_kyber512" : {"avx2_keypair" : "kyberfivekey", "avx2_enc" : "kyberfiveenc", "avx2_dec" : "kyberfivedec"},
            "anon_kem_kyber_kyber768" : {"avx2_keypair" : "kybersevenkey", "avx2_enc" : "kybersevenenc", "avx2_dec" : "kybersevendec"},
            "anon_scalarmult_curve25519" : {"ref4" : "curvereffour", "ref4_base" : "curvereffourbase", "ref5" : "curvereffive",
                                            "ref5_base" : "curvereffivebase", "mulx" : "curvemulx", "mulx_base" : "curvemulxbase" }
        }
cols = ["a", "b", "c"]

colmean = {"a" : "Unprotected cpucycles", "b" : "Protected cpucycles", "c" : "performance overhead"}

implname = {
              "anon_kem_kyber_kyber512" : "kyber512",
              "anon_kem_kyber_kyber768" : "kyber768",
              "anon_scalarmult_curve25519" : "curve25519"
}


def main():
    result = ""
    ref_dict = {}

    parser = argparse.ArgumentParser()
    parser.add_argument('--indir', default="main-sslh", type=str)
    parser.add_argument('--outfile', default="results1_data.tex", type=str)
    args = parser.parse_args()
    indir = args.indir
    outfile = args.outfile

    for impl in impls1:
        for func in impls1[impl]:
            inputfile = "%s/%s_amd64_%s.csv" % (indir, impl, func)
            with open(inputfile) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                for row in csv_reader:
                    for i in range(len(row)):
                        data = r"\newcommand*\%s%s{%s}" % (impls1[impl][func], cols[i], row[i])
                        data += "\t\t\t% "
                        data += r"%s for %s %s" % (colmean[cols[i]], implname[impl], func)
                        # print(data)
                        result += data
                        result += "\n"
                result += "\n"
        result += "\n\n"
    # print(result)

    with open(outfile, 'w') as f:
        f.write(result)

if __name__ == "__main__":
    main()
