#!/usr/bin/env python3

import os
import csv
import sys
import argparse

impls1 = {
            "anon_stream_chacha_chacha12" : {"ref" : "chachaoneref", "ref_xor" : "chachaonerefxor", "avx" : "chachaoneavx",
                                            "avx_xor" : "chachaoneavxxor", "avx2" : "chachaoneavxtwo", "avx2_xor" : "chachaoneavxtwoxor" },
            "anon_stream_chacha_chacha20" : {"ref" : "chachatworef", "ref_xor" : "chachatworefxor", "avx" : "chachatwoavx",
                                            "avx_xor" : "chachatwoavxxor", "avx2" : "chachatwoavxtwo", "avx2_xor" : "chachatwoavxtwoxor" },
            "anon_stream_chacha_chacha20_ietf" : {"ref" : "chachaietfref", "ref_xor" : "chachaietfrefxor", "avx" : "chachaietfavx",
                                            "avx_xor" : "chachaietfavxxor", "avx2" : "chachaietfavxtwo", "avx2_xor" : "chachaietfavxtwoxor" },
            "anon_stream_salsa20_salsa20" : {"ref" : "salsatworef", "ref_xor" : "salsatworefxor", "avx" : "salsatwoavx",
                                            "avx_xor" : "salsatwoavxxor", "avx2" : "salsatwoavxtwo", "avx2_xor" : "salsatwoavxtwoxor" },
            "anon_stream_salsa20_salsa2012" : {"ref" : "salsatwolfref", "ref_xor" : "salsatwolfrefxor", "avx" : "salsatwolfavx",
                                            "avx_xor" : "salsatwolfavxxor", "avx2" : "salsatwolfavxtwo", "avx2_xor" : "salsatwolfavxtwoxor" },
            "anon_stream_xsalsa20" : {"ref" : "xsalsatworef", "ref_xor" : "xsalsatworefxor", "avx" : "xsalsatwoavx",
                                            "avx_xor" : "xsalsatwoavxxor", "avx2" : "xsalsatwoavxtwo", "avx2_xor" : "xsalsatwoavxtwoxor" },
            "anon_onetimeauth_poly1305" : { "ref" : "polyref", "ref_verify" : "polyrefverify", "avx" : "polyavx", "avx_verify" : "polyavxverify",
                                            "avx2" : "polyavxtwo", "avx2_verify" : "polyavxtwoverify"},
            "anon_secretbox_xsalsa20poly1305" : { "ref" : "xsalsapolyref", "ref_open" : "xsalsapolyrefopen", "ref_open_forgery" : "xsalsapolyrefopenforgery",
                                                    "avx" : "xsalsapolyavx", "avx_open" : "xsalsapolyavxopen", "avx_open_forgery" : "xsalsapolyavxopenforgery",
                                                    "avx2" : "xsalsapolyavxtwo", "avx2_open" : "xsalsapolyavxtwoopen", "avx2_open_forgery" : "xsalsapolyavxtwoopenforgery" },
            "anon_hash_sha3_224" : { "ref" : "shatttfref", "avx2" : "shatttfavxtwo"},
            "anon_hash_sha3_256" : { "ref" : "shattfsref", "avx2" : "shattfsavxtwo"},
            "anon_hash_sha3_384" : { "ref" : "shattefref", "avx2" : "shattefavxtwo"},
            "anon_hash_sha3_512" : { "ref" : "shatfotref", "avx2" : "shatfotavxtwo"},
            "anon_hash_sha256" : { "ref" : "shatfsref"}


        }
cols = ["x", "y", "z"]
inlen = {"32" : "a", "128" : "b", "1024" : "c", "16384" : "d"}

implname = {
    "anon_stream_chacha_chacha12" : "chacha12",
    "anon_stream_chacha_chacha20" : "chacha20",
    "anon_stream_chacha_chacha20_ietf" : "chacha20_ietf",
    "anon_stream_salsa20_salsa20" : "salsa20",
    "anon_stream_salsa20_salsa2012" : "salsa2012",
    "anon_stream_xsalsa20" : "xsalsa20",
    "anon_onetimeauth_poly1305" : "poly1305",
    "anon_secretbox_xsalsa20poly1305" : "xsalsa20poly1305",
    "anon_hash_sha3_224" : "sha3_224",
    "anon_hash_sha3_256" : "sha3_256",
    "anon_hash_sha3_384" : "sha3_384",
    "anon_hash_sha3_512" : "sha3_512",
    "anon_hash_sha256" : "sha256",
}
colmean = {"x" : "Unprotected cpucycles", "y" : "Protected cpucycles", "z" : "performance overhead"}

def main():
    result = ""
    ref_dict = {}

    parser = argparse.ArgumentParser()
    parser.add_argument('--indir', default="main-sslh", type=str)
    parser.add_argument('--outfile', default="results2_data.tex", type=str)
    args = parser.parse_args()
    indir = args.indir
    outfile = args.outfile

    for impl in impls1:
        for func in impls1[impl]:
            inputfile = "%s/%s_amd64_%s.csv" % (indir, impl, func)
            with open(inputfile) as csv_file:
                csv_reader = csv.reader(csv_file, delimiter=',')
                for row in csv_reader:
                    if row[0] in inlen:
                        for i in range(1,  len(row)):
                            data = r"\newcommand*\%s%s%s{%s}" % (impls1[impl][func], inlen[row[0]], cols[i - 1], row[i])
                            data += "\t\t\t% "
                            data += r"%s for %s %s of %s bytes" % (colmean[cols[i - 1]], implname[impl], func, row[0])
                            result += data
                            result += "\n"
                    result += "\n"
        result += "\n\n"
    # print(result)

    with open(outfile, 'w') as f:
        f.write(result)

if __name__ == "__main__":
    main()
